/*
 * packetmodeller/packet_modeller.c
 * $Rev: 846 $
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/**
 * @file packet_modeller.c
 * @brief PacketModeller - Traffic redirection deamon for virtual wireless devices
 *
 * Usage: packetmodeller MODELIP:MODELPORT LOCALPORT
 *                       [-i INTERFACE][-6][-n][-d]
 *
 * Tunnels traffic from a virtual wireless interface to a
 * wireless emulation model. To create the virtual interface
 * start `vifd` first. The virtual interfaces are then detected
 * automatically.
 *
 * Parameters:
 *
 * MODELIP:PORT IP Address/Hostname and port number of wireless model
 * LOCALPORT    Local port to listen for model replies
 * -6           Use IPv6 protocol
 * -d           Enable debug output
 * -h           Show this help message
 * -i INTERFACE	Tunnel interface (default: eth0)
 * -n           Enable node-to-node communication (only for testing!)
 */

/* INCLUDES *******************************************************************/

#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <stdarg.h>                 // va_list, va_start(), va_end(), vfprintf(), ...
#include <stdio.h>                  // printf(), fprintf()
#include <stdlib.h>                 // exit()
#include <string.h>                 // memset(), strcpy(), ...
#include <syslog.h>                 // openlog(), vsyslog(), LOG_PID, ...
#include <unistd.h>                 // getopt(), optarg
#include <vif_notify.h>             // vif_notify_poll()
#include "netdevice.h"              // init_udp_socket()
#include "message.h"
#include "packet_modeller.h"
#include "pm_print.h"


/* EXTERNAL FUNCTIONS *********************************************************/

extern int pm_init_host();
extern int pm_init_model(const char *);
extern int pm_init_net();
extern int pm_model_register();
extern int pm_model_deregister();
extern void pm_model_update(int, const char*);


/* PRIVATE VARIABLES **********************************************************/

char *exec_name = NULL;                     /**< Program name */
const char pm_version[] = "0.2.0beta\0";    /**< Program version */

/* Host parameters */
pm_client_info client_info;                 /**< Wireless client information */
pm_model_info model_info;                   /**< Wireless model information */

/** Number of virtual interfaces */
int num_vif;

/* Socket file descriptors */
int tap_sockets[MAX_NUM_VIF];               /**< TAP socket file descriptors */
int net_socket;                             /**< Network socket file descriptor */

/** Indicate direct node to node mode */
int direct;

/** Are we registered at the WlanModel */
int is_registered;

/** IP protocol version to use (AF_INET for IPv4, AF_INET6 for IPv6) */
int ip_family;


/* FUNCTIONS *********&********************************************************/

/** Print usage details */
void
pm_usage()
{
    printf("\nUsage: %s MODELIP:MODELPORT LOCALPORT\n", exec_name);
    printf("\t\t\t[-i INTERFACE][-6][-n][-d]\n");
    printf("Try `%s -h' for more details\n", exec_name);

    exit(EXIT_FAILURE);
}

/** Print help */
void
pm_help()
{
    printf("\nUsage: %s MODELIP:MODELPORT LOCALPORT\n", exec_name);
    printf("\t\t\t[-i INTERFACE][-6][-n][-d][-h]\n");
    printf("\nTunnels traffic from a virtual wireless interface to a\n");
    printf("wireless emulation model. To create the virtual interface\n");
    printf("start `vifd` first. The virtual interfaces are then detected\n");
    printf("automatically.\n");
    printf("\nParameters:\n");
    printf("MODELIP:PORT\tIP Address/Hostname and port number of wireless model\n");
    printf("LOCALPORT\tLocal port to listen for model replies\n");
    printf("-6\t\tUse IPv6 protocol\n");
    printf("-d\t\tEnable debug output\n");
    printf("-h\t\tShow this help message\n");
    printf("-i INTERFACE\tTunnel interface (default: %s)\n", client_info.host_if);
    printf("-n\t\tEnable node-to-node communication (only for testing!)\n");

    exit(EXIT_SUCCESS);
}

/** Stub for print_debug signature */
int
ignore_debug(const char* text, ...)
{
    return 0;
}

/** Print debug message with variable argument list */
int
print_debug(const char *format, ...)
{
    va_list arg;

    va_start(arg, format);
    pm_debug_args(format, arg);
    va_end(arg);

    return 0;
}

/** Stub for print_debug_args signature */
void
ignore_debug_args(const char *format, va_list arg)
{
}

/** Print debug message with given argument list */
void
print_debug_args(const char *format, va_list arg)
{
    FILE *stdout = fopen("/dev/stdout", "a");
    fprintf(stdout, "DEBUG: ");
    vfprintf(stdout, format, arg);
    fclose(stdout);
}

/** Print error message */
int
pm_error(const char *format, ...)
{
    int chars_printed = 0;
    va_list arg;

    FILE *stderr = fopen("/dev/stderr", "a");
    if (stderr == NULL)
        return -1;
    chars_printed = fprintf(stderr, "Error: ");
    va_start(arg, format);
    chars_printed += vfprintf(stderr, format, arg);
    va_end(arg);
    fclose(stderr);

    return chars_printed;
}

/** Send message to syslog and print to debug output */
void
pm_log(const char *format, ...)
{
    va_list arg;

    // send to syslog
    openlog(exec_name, LOG_PID, LOG_SYSLOG);
    va_start(arg, format);
    vsyslog(LOG_INFO, format, arg);
    va_end(arg);
    closelog();

    // if we enabled debug, also print to stdout
    va_start(arg, format);
    pm_debug_args(format, arg);
    va_end(arg);
}

/** Shutdown packetmodeller and deregister at WlanModel */
void
pm_shutdown(int signal)
{
    int i;
    int rtrn = 0;

    // deregister model
    if (is_registered == 1)
        if ((rtrn != pm_model_deregister()) < 0);
            signal = rtrn;

    // close network sockets
    close(net_socket);
    for (i = 0; i < num_vif; i++)
        close(tap_sockets[i]);

    if (((signal == SIGINT) ||
        (signal == SIGABRT) ||
        (signal == SIGTERM)) &&
        (rtrn == 0))
    {
        exit(EXIT_SUCCESS);
    } else {
        exit(signal);
    }
}


/** Listen to TUN and network interfaces for "wireless" packets */
void
pm_daemon()
{
    fd_set fds;
    int i, recv_bytes, sent_bytes;
    register int ret;
    char buffer[DATA_MSG_SIZE];                     // biggest possible message
    struct data_msg *message = (struct data_msg *) buffer;
    raw_packet *ether_frame = (raw_packet *) message->wlan_pkt;

    pm_debug("pm_daemon(): Running main loop with %i vif(s)\n", num_vif);
    pm_debug("\n");

    // find biggest fd, for select()
    int max_fd = 0;
    for (i = 0; i < num_vif; i++)
        max_fd = max(max_fd, tap_sockets[i]);
    max_fd = max(max_fd, net_socket) + 1;

    pm_debug(" net_socket: fd: %i\n", net_socket);
    pm_debug(" vif_index: 0, fd: %i\n", tap_sockets[0]);
    pm_debug(" vif_index: 1, fd: %i\n", tap_sockets[1]);
    pm_debug("\n");

    for (;;)
    {
        FD_ZERO(&fds);
        for (i = 0; i < num_vif; i++)
            FD_SET(tap_sockets[i], &fds);
        FD_SET(net_socket, &fds);

        // check for fd with pending input data
        if ((ret = select(max_fd, &fds, NULL, NULL, NULL)) < 0)
        {
            pm_error("pm_daemon(): select() (%s)\n",
                    strerror(errno));
            return;
        } else {
            // check tap device for incoming traffic
            for (i = 0; i < num_vif; i++)
            {
                if (FD_ISSET(tap_sockets[i], &fds))
                {
                    // read packet into wlan_pkt argument of wlan_msg
                    recv_bytes = read(tap_sockets[i], ether_frame,
                                    MAX_FRAME_SIZE);
                    pm_debug(">>tap_fd<< %i Bytes read\n", recv_bytes);
                    if (recv_bytes > 0)
                    {
                        message->header.type = DATA_MSG;
                        memmove(&message->header.id, client_info.host_id,
                                sizeof(client_info.host_id));
                        message->pkt_length = recv_bytes;
                        message->vif_index = i;
                        debug_msg((char *) message);

                        // send data to model
                        if ((sent_bytes = sendto(net_socket, message,
                                DATA_PREAMBLE_SIZE + recv_bytes, 0,
                                model_info.host_info.addr_info.ai_addr,
                                model_info.host_info.addr_info.ai_addrlen)) < 0)
                        {
                            pm_error("pm_daemon(): net:sendto (%s)\n",
                                    strerror(errno));
                        }
                        pm_debug(">>net_socket<< %i Bytes sent\n", sent_bytes);
                        pm_debug("\n");
                    } else
                        pm_error("pm_daemon(): tap:read (%s)\n", strerror(errno));
                }
            }
            // check network device for incoming traffic
            if (FD_ISSET(net_socket, &fds))
            {
                recv_bytes = recvfrom(net_socket, buffer, DATA_MSG_SIZE, 0,
                                    NULL, NULL);
                pm_debug(">>net_socket<< %i Bytes read\n", recv_bytes);
                if (recv_bytes > 0)
                {
                    if (is_data_msg((struct msg_header *) buffer))
                    {
                        debug_msg(buffer);

                        // send to TAP interface
                        pm_debug("pm_daemon(): sendto fd = %i, pkt_length = %i\n",
                                tap_sockets[message->vif_index],
                                message->pkt_length);
                        if ((sent_bytes = write(
                                tap_sockets[message->vif_index],
                                ether_frame, message->pkt_length)) < 1)
                        {                       // handle write() error
                            pm_error("pm_daemon(): write() (%s)\n",
                                    strerror(errno));
                        } else {
                            pm_debug(">>tap_fd<< %i Bytes sent\n", sent_bytes);
                            pm_debug("\n");
                        }
                    } else {
                        pm_debug("pm_daemon(): Unknown message received\n");
                        pm_debug("\n");
                    }
                } else                          // handle recvfrom() errors
                    pm_error("pm_daemon(): recvfrom() (%s)\n", strerror(errno));
            }
        }
    }
}


/** Parse command line options and checks it for validity */
int
pm_pars_opts(int argc, char *argv[])
{
    int opt = -1;
    int debug = 0;
    int req_args = 3;
    int add_args = 0; // required because getopts changes the order of arguments
    int port_conv = -1;

    // IMPORTANT: Don't use pm_log or pm_debug before here!
    pm_debug = ignore_debug;
    pm_debug_args = ignore_debug_args;

    while ((opt = getopt(argc, argv, "6dnhi:")) != -1)
    {
        switch (opt)
        {
            case 'h':
                pm_help();
                break;
            case 'i':
                add_args += 2;
                break;
            case 'd':
                add_args++;
                debug = 1;
                break;
            case 'n':
                add_args++;
                break;
            case '6':
                add_args++;
                ip_family = PF_INET6;
                break;
        }
    }
    optind = 0; // reset getopt execution

    if (argc < (req_args + add_args))
    {
        pm_error("Missing arguments\n");
        return -1;
    }

    // configure debug/error streams
    if (debug == 1)
    {
        pm_debug = print_debug;
        pm_debug_args = print_debug_args;
        vif_set_debug_function(pm_debug);
        vif_notify_set_debug_function(pm_debug);
    }
    vif_set_error_function(pm_error);
    vif_notify_set_error_function(pm_error);

    pm_log("Starting PacketModeller %s\n", &pm_version);
    if (debug == 1)
        pm_log("Debugging enabled!\n");
    if (ip_family == AF_INET6)
        pm_log("Using IPv6 protocol\n");

    // create a channel to the NET kernel
    if (init_udp_socket() < 0)
        return (EXIT_FAILURE);

    // read model address
    if (pm_init_model(argv[1 + add_args]) == -1)
    {
        pm_error("Model initialization failed!\n");
        return -1;
    }

    // read local port
    port_conv = atoi(argv[2 + add_args]);  // first store in normal int to check range
    if ((port_conv > 65536) || (port_conv < 1))
    {
        pm_error("Invalid local port number\n");
        return -1;
    } else {
        memmove(client_info.host_port, argv[2 + add_args], 5);
    }

    while ((opt = getopt(argc, argv, "6dni:")) != -1)
    {
        switch (opt)
        {
            case 'n':
                direct = 1;
                break;
            case 'i':
                if (get_ip_by_dev(NULL, optarg)) // check if configured interface
                    return -1;
                else
                {   // we cannot trust the input, correctly create string
                    strncpy(client_info.host_if, optarg, IFNAMSIZ);
                    client_info.host_if[IFNAMSIZ] = '\0';
                }
                break;
        }
    }
    pm_debug("Using local interface %s\n", client_info.host_if);

    return 0;
}

/** MAIN */
int
main(int argc, char *argv[])
{
    pthread_t tid;

    // extract executable name
    if (strrchr(argv[0], '/') == NULL)
    {
        exec_name = argv[0];
    } else {
        exec_name = strrchr(argv[0], '/') + 1;
    }

    // initialize default values
    num_vif = 0;
    ip_family = AF_INET;                    // AF_INET (ipv4), AF_INET6 (ipv6)
    direct = 0;
    is_registered = 0;
    memset(&model_info, 0, sizeof(pm_model_info));
    memset(&client_info, 0, sizeof(pm_client_info));
    strcpy(client_info.host_if, "eth0\0");            // set default interface

    // check parameters
    if (pm_pars_opts(argc, argv) < 0)
        pm_usage();

    // initialize host values
    if (pm_init_host() < 0)
    {
        pm_error("Host initialization failed!\n");
        return EXIT_FAILURE;
    }

    // registering signal handler
    signal(SIGINT, pm_shutdown);
    signal(SIGABRT, pm_shutdown);
    signal(SIGTERM, pm_shutdown);

    // initialize network sockets
    if (pm_init_net() < 0)
    {
        pm_error("Network initialization failed!\n");
        return EXIT_FAILURE;
    }

    pm_debug("Successfully finished host initialization!\n");
    pm_debug("\n");

    // register node at WlanModel
    if (direct != 1)
    {
        if ((is_registered = pm_model_register()) != 1)
            return EXIT_FAILURE;
    } else {
        pm_log("Argument '-n' found: Not registering host!\n");
    }

    // create thread for updating model at vif configuration changes
    if (pthread_create(&tid, NULL, vif_notify_poll,
        (void *) &pm_model_update) < 0)
    {
        pm_error("Couldn't create new thread (%s)\n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    // run packet redirecting daemon
    pm_daemon();

    exit(EXIT_SUCCESS);
}
